#include <stdio.h>
#include <stdlib.h>

/*#define DEBUGGERY   /*displays various debug messages*/
/*#define FILE_DEBUG  /*outputs a permanant, not temporary, file*/
#define INTEL       /*assumes your PC uses INTEL formatting*/

/*Macros*/
#define error(S)	fprintf(stderr,S)

/*error messages and whatnot*/
int erratta(int code)
{switch(code)
	{case 0: error("\nAn unknown non-fatal error occurred."); break;
	case 1:  error("\nAn unknown fatal error occurred!  Aegh!"); break;
	case 2:  error("\nUnable to open a temporary file."); break;
	case 3:  error("\nTerrifying buffer overrun.  Mention that to me!"); break;
	/*remove when EDL1 is finally added*/
	case 100: case 101: case 102:
         fprintf(stderr,"\nOkay, I lied. EDL%i hasn't actually been implemented yet...\n\tSorry for the hassle.",code-100); break;
      }
 if(code) exit(code);
 return code;}

/*standard 32bit byteswapping*/
unsigned long byteswap(unsigned long w)
{return (w >> 24) | ((w >> 8) & 0x0000ff00) | ((w << 8) & 0x00ff0000) | (w << 24);}    

/*returns file size*/
unsigned long filesize(FILE *in)
{fseek(in,0,SEEK_END);
 return ftell(in);}

/*file copying routine: copies 1000 bytes of blank data to temp, then the original file contents
  mode 0 just copies; mode 1+ adds the blank data*/
unsigned long filecopy(FILE *in, FILE *out,int mode)
{unsigned long x,y;

 /*Copy in 0x1000  of zeroes.  LZ assumes anything prior to the file is zero, 
 and the N64 does this to the file prior to decompression*/
 if(mode) {for(x=0;x<0x1000;x++) fputc(0,out);}
 
 /*copy the original file's contents to end of buffer*/
 x=filesize(in);
 fseek(in,0,SEEK_SET);
 for(y=0;y<x;y++) fputc(fgetc(in),out);
 
 return ftell(out);}

/*this takes a [value] of [size] bits and inserts it into [data], incrementing [bitcount]
  returns the bitcount*/
unsigned long helper(unsigned long *data,unsigned long *bitcount,unsigned long value,int size,FILE *out,unsigned long *pos)
{unsigned long v,w,x,y,z;
 /*shift current data and tack on new value*/
 x=*data;
 y=*bitcount;
 z=*pos;
 
#ifdef DEBUGGERY
 printf("\nBuffer: %08X-%X\tvalue: %X\tsize: %X",x,y,value,size);
#endif

 /*copy the value one bit at a time to the data buffer to prevent overruns*/
for(w=0;w<size;w++)
      {v=value>>w;       /*bit at value*/
       v&=1;
       x|=v<<y;          /*bit is pushed bitcount bits left and tacked onto value*/
       /*update bitcount and test for overrun*/
       y++;
       if(y>32) erratta(3);
       /*copy the buffer to file if it happens to equal 32bits*/
       if(y<32) continue;
       fseek(out,z,SEEK_SET);
#ifdef INTEL
       x=byteswap(x);
#endif
       fwrite(&x,4,1,out);
       x=0;
       y=0;
       z+=4;
       }

#ifdef DEBUGGERY
 printf("\tnFinal: %08X-%X",x,y);
#endif 
 *data=x;
 *bitcount=y;
 *pos=z;
 return y;}


/*EDL compression types.
  All accept the copied input file, an output file, and return the compressed file size*/
/*EDL compression type 0 - store
  no routine required.  It is store, after all...*/

/*EDL compression type 1 - LZ*/
unsigned long EDLcmp1(FILE *in,FILE *out)
{unsigned long x,ipos=0x1000,opos=12;
 /*this is a silly way to get the compressed file size, but it does work*/
 return ftell(out);}

/*EDL compression type 2 - RLE*/
/*uses a series of bits to determine what each run consists of
0	write next byte
110	num=2
1110	num=3,  offset
1111xxxxxxxx	num=x, offset
1000	num=4, offset
1010	num=5, offset
10010	num=6, offset
10011	num=7, offset
10110	num=8, offset
10111xxxx	write next x+3 words	(x+3)*4 bytes
111100000000	terminate

offset types:
	0	0x0
	110	0x100
	1000	0x200
	1001	0x300
	10101	0x400
	10111	0x500
	11101	0x600
	11111	0x700
	101000	0x800
	101001	0x900
	101100	0xA00
	101101	0xB00
	111000	0xC00
	111001	0xD00
	111100	0xE00
	111101	0xF00

backtrack:
xxxxxxxx	backtrack value (added to offset above when applicable)
reads num bytes from backtrack+offset

data retains the bits that will be copied every 16bits.  bitcount goes up as you write bits to it.

buffer stores a record of each hit for a match, then go back through to determine which is best
	index keeps the number of entries in buffer[]
string is your search string of len size, compared against file.  len=n in strncmp
Idea is you fill window with first 0x2000 bytes.  
	Load two bytes into search and find all matches between here and its position
	then, get next byte and search all these locations again
	repeat until the largest possible string is found, and use the one occurring at the closest address
search string, at most, can be 256 bytes, since largest type is num=FF. 
	longest possible straight data write is 0x48 bytes - 0x12 words

only stupid thing has to do with a run of straight written data
	if  a match can't be found, increment run
		when a match finally is found, first you copy in the entire run, set run to zero, then do the len
[number] dictates the maximum length of the search[] string*/
unsigned long EDLcmp2(FILE *in,FILE *out)
{unsigned long x,y,size,data=0,bitcount=0,ipos=0x1000,opos=12,index=0,len=0,run=0,number=0x100;
 unsigned short buffer[0x1000];
 unsigned char search[256],window[0x1100];

#ifdef DEBUGGERY
 printf("\nStarting EDL2 compression...");
#endif

 size=filesize(in);
 if((size-ipos)<number) number=size-ipos;
 
 /*initially fill window - later, you need only move it*/
 fseek(in,0,SEEK_SET);
 for(x=0;x<0x1100;x++) window[x]=fgetc(in);
 /*now fill the search window with max characters - again, will move it later*/
 fseek(in,ipos,SEEK_SET);
 for(x=0;x<number;x++) search[x]=fgetc(in);
 
 len=0;
 while(ipos+len<size)
	{/*fix ipos, refill the buffers, then loop*/
	 ipos+=len;
     for(x=0;x<0x1000;x++) window[x]=window[x+len];
     if((size-ipos)<number)
       {number=size-ipos;
#ifdef DEBUGGERY
       printf("\nSet number to %X-%X=%X",size,ipos,number);
#endif
        }
	 for(x=0;x<number;x++)
	    {if((x+len)<number) search[x]=search[x+len];
         else
           {fseek(in,ipos+x,SEEK_SET);
	        search[x]=fgetc(in);
            }
	     window[x+0x1000]=search[x];
	     }
#ifdef DEBUGGERY
       if(number<0x8)
         {printf("\n\t");
         for(x=0;x<number;x++) printf("%02X ",search[x]);
         }
#endif
     
     index=0;
	 len=2;
	 /*search for a hit of len+2*/
     for(x=0;x<0x1000;x++)
        {/*as you find them, record the positions in buffer and increment index*/
         if(!memcmp(window+x,search,len))
           {buffer[index]=(short)x;    /*evil pointer math*/
            index++;
            }
         }/*end for()*/
	 /*if you found matches (index>0), inc len, then go back through them testing for len+2 entries*/
     if(index && number>1)
       {do {y=0;
            for(x=0;x<index;x++)
               {if(!memcmp(window+buffer[x],search,len+1))
                  {buffer[y]=buffer[x];
                   y++;
                   }/*end if(onhit)*/
               }/*end for()*/
            if(y) {index=y;
                   len++;
                   }
           if(len==0xFF || len>number) break;        /*kill if you hit the max string length*/
            } while(y);
#ifdef DEBUGGERY
	       printf("\n\tMatches! len: %X\thits: %X",len,index);
#endif
           /*now that you have a match, find the favorite match and set index to this value*/
	       y=0;
           for(x=0;x<index;x++)
              {if(buffer[y]<buffer[x]) y=x;
               }
           index=y;
#ifdef DEBUGGERY
           printf("\tusing: #%X-%04X %04X\n\t",index,buffer[index],0xFFF-buffer[index]);
	       for(x=0;x<len;x++) printf("%02X ",search[x]);
#endif
           /*convert happy buffer value to backward offset*/
           if(buffer[index]>=0x1000) len=1;
           else buffer[index]=0xFFF-buffer[index];
	       /*bugfix: kill len2 entries that fall outside the first 0x100
             also kill stupid buffer values, like <=0*/
           if(len==2 && buffer[index]>0xFF) len=1;
        }/*end if(index)*/
	 else len=1;  /*straight binary run*/

     /*when len=1 this indicates a run
       normally, just increment run
       however, near EOF you'll need to write them*/
     if(len<2)
       {run++;
        if((ipos+len)<size) continue;
        ipos++;
        }
	 
	 /*first, test if run>0.  If so, you first have to copy in the run of straight characters*/
	 if(run)
		{/*copy any sets of words first - start at ipos-run*/
#ifdef DEBUGGERY
        printf("\n  Copying run of %X chars at %X",run,ipos-run-0x1000);
#endif
		fseek(in,ipos-run,SEEK_SET);
		x=run/4;
		while(x>3)
			{helper(&data,&bitcount,0x1D,5,out,&opos);	/*command type*/
			 x-=3;
			 y=x&0xF;
			 helper(&data,&bitcount,y,4,out,&opos);		/*copy size to command*/
			 x-=y;
			 y=(y+3)*4;
			 for(run-=y;y>0;y--)			/*copy actual data*/
				{helper(&data,&bitcount,fgetc(in),8,out,&opos);
				 }
			 }
		/*write individual bytes, then set run to zero*/
		while(run>0)
			{helper(&data,&bitcount,0,1,out,&opos);
			 helper(&data,&bitcount,fgetc(in),8,out,&opos);
			 run--;
			 }
		}
	 /*crap, will need to figure out a way to loop for copies, do a run, and not break on EOF*/
	 if(len<2) continue;

	 /*do the copy byte commands*/
	 switch(len)
		{case 2:
			helper(&data,&bitcount,0x3,3,out,&opos);
			break;
		 case 3:
			helper(&data,&bitcount,0x7,4,out,&opos);
			break;
		 case 4:
			helper(&data,&bitcount,0x1,4,out,&opos);
			break;
		 case 5:
			helper(&data,&bitcount,0x5,4,out,&opos);
			break;
		 case 6:
			helper(&data,&bitcount,0x9,5,out,&opos);
			break;
		 case 7:
			helper(&data,&bitcount,0x19,5,out,&opos);
			break;
		 case 8:
			helper(&data,&bitcount,0xD,5,out,&opos);
			break;
		 default:
			helper(&data,&bitcount,0xF,4,out,&opos);
			helper(&data,&bitcount,len-8,8,out,&opos);
			break;
		}
	 
     /*in every case but 2, also send a sinister offset chunk*/
	 if(len>2)
		{     if(buffer[index]>=0xF00) helper(&data,&bitcount,0x2F,6,out,&opos);
		 else if(buffer[index]>=0xE00) helper(&data,&bitcount,0x0F,6,out,&opos);
		 else if(buffer[index]>=0xD00) helper(&data,&bitcount,0x27,6,out,&opos);
		 else if(buffer[index]>=0xC00) helper(&data,&bitcount,0x07,6,out,&opos);
		 else if(buffer[index]>=0xB00) helper(&data,&bitcount,0x2D,6,out,&opos);
		 else if(buffer[index]>=0xA00) helper(&data,&bitcount,0x0D,6,out,&opos);
		 else if(buffer[index]>=0x900) helper(&data,&bitcount,0x25,6,out,&opos);
		 else if(buffer[index]>=0x800) helper(&data,&bitcount,0x05,6,out,&opos);
		 else if(buffer[index]>=0x700) helper(&data,&bitcount,0x1F,5,out,&opos);
		 else if(buffer[index]>=0x600) helper(&data,&bitcount,0x17,5,out,&opos);
		 else if(buffer[index]>=0x500) helper(&data,&bitcount,0x1D,5,out,&opos);
		 else if(buffer[index]>=0x400) helper(&data,&bitcount,0x15,5,out,&opos);
		 else if(buffer[index]>=0x300) helper(&data,&bitcount,0x09,4,out,&opos);
		 else if(buffer[index]>=0x200) helper(&data,&bitcount,0x01,4,out,&opos);
		 else if(buffer[index]>=0x100) helper(&data,&bitcount,0x03,3,out,&opos);
		 else 				 helper(&data,&bitcount,0,1,out,&opos);
		}
	 /*set backtrack value*/
	 helper(&data,&bitcount,buffer[index]&0xFF,8,out,&opos);
#ifdef DEBUGGERY
	 printf("\n  Wrote backtrack @ %X: %X, %X long -%X bytes back",opos,ipos-0x1000,len,buffer[index]);
#endif
     }
 
 /*set the terminator at end of file*/
 helper(&data,&bitcount,0x0F,12,out,&opos);	/*command type*/
 /*bugfix: flush the buffer with a few calls of 0*/
 if(bitcount) helper(&data,&bitcount,0,32-(bitcount%32),out,&opos);
 
 /*this is a silly way to get the compressed file size, but it does work*/
 return opos;}


/*main loop
  open files and redirect to compression routines*/
int main(int argc, char *argv[])
{char filename[130];
 FILE *in, *out, *temp;
 int numin=0,numout=0,type=2,endian=0x80;
 unsigned long x,y;
 
 for(x=1;x<argc;x++)
    {if(argv[x][0]=='-' || argv[x][0]=='/')
      {switch(argv[x][1])
		{case 'O': case 'o': numout=x; break;
		 case 'T': case 't': 
			strcpy(filename,argv[x]+2);
			if(!strncasecmp(filename,"store",5) || !strncmp(filename,"0",1)) type=0;
			else if(!strncasecmp(filename,"LZ",2) || !strncmp(filename,"1",1)) type=1;
/*			else if(!strncasecmp(filename,"RLE",3) || !strncmp(filename,"2",1)) type=2;*/
			else type=2;
			break;
		 case 'E': case 'e': 
			strcpy(filename,argv[x]+2);
			if(!strncasecmp(filename,"little",5) || !strncmp(filename,"0",1)) endian=0;
/*			else if(!strncasecmp(filename,"big",2) || !strncmp(filename,"1",1)) endian=0x80;*/
			else endian=0x80;
			break;
		 case 'H': case 'h': case '?':
                  printf("\nEDL 0-2 Compressor for Eurocom N64 Games");
                  printf("\nUsage: %s [filename] -O[filename] -T[#] -E[#] -H",basename(argv[0]));
			printf("\n\tOptions can be written with either - or /");
			printf("\n\t[filename]\tOptional input filename");
			printf("\n\t-O[filename]\tOptional output filename");
			printf("\n\t-T[#]\tCompression type: 0(store) ,1(LZ) ,2(RLE)");
			printf("\n\t-E[#]\tEndianess: 0(little), 1(big)");
			printf("\n\t-H or -?\tThis help message");
			return 0;
              }
      }
    else numin=x;
    }

/*remove when the file format conversions are finished*/
if(type==1) erratta(101);

#ifdef FILE_DEBUG
 if(!(temp=fopen("debug.bin","wb+"))) erratta(2);
#else
 if(!(temp=tmpfile())) erratta(2);
#endif

 strcpy(filename,"\0");
 if(numin) strcpy(filename,argv[numin]);
 while (!(in = fopen(filename, "rb"))) {
	printf("\nInput binary file? ");
	strcpy(filename,"\0");	
	gets(filename);
    }

 if(numout) strcpy(filename,argv[numout]+2);
 else strcat(filename,".edl");
 while (!(out = fopen(filename, "wb+"))) {
	printf("\nOutput EDL file? (.edl)");
	strcpy(filename,"\0");
	gets(filename);
    }

 /*tack on an EDL header - INTEL order
   x will be decompressed size and the first four bytes of filename=type
   compressed size is a placeholder, technically*/
 fseek(out,0,SEEK_SET);
 sprintf(filename,"EDL\0\0\0\0\0");
 filename[3]=endian|type;
 x=filesize(in);
 fwrite(filename,1,8,out);
#ifdef INTEL
 if(endian)  x=byteswap(x);
#else
 if(!endian) x=byteswap(x);
#endif
 fwrite(&x,4,1,out); 

 /*create a copy of the original binary with a prelude of 0x1000 zeroes; returns filesize*/
 /*EDLcmp0 - filecopy*/
 if(!type) y=filecopy(in,out,type);
 else filecopy(in,temp,type);
 fclose(in);

 /*start compression*/
 if(type==2) 	  y=EDLcmp2(temp,out);
 else if(type==1) y=EDLcmp1(temp,out);
 /*update compressed size and close files*/
 fseek(out,4,SEEK_SET);
 x=y;
#ifdef INTEL
 if(endian)  y=byteswap(y);
#else
 if(!endian) y=byteswap(y);
#endif
 fwrite(&y,4,1,out);
 fclose(temp);
 
 /*display compression ratio and stuff*/
 fseek(out,8,SEEK_SET);
 fread(&y,4,1,out);
#ifdef INTEL
 if(endian)  y=byteswap(y);
#else
 if(!endian) y=byteswap(y);
#endif 
 x-=12;
 printf("\nFinished.\t");
 if(!type) printf("Store used, so no compression occurred.\n");
 else	   printf("%s compression used; compressed %i%% (%X to %X).\n",type<2 ? "LZ":"RLE",(int)(x*100/y),y,x);
 fclose(out);
 return 0;}
